"""Text User Interface (TUI) utilities"""

from __future__ import annotations

import logging
import re
from collections.abc import Callable
from typing import Any

logger = logging.getLogger(__name__)


class Prompts:
    """`input()` builtin on steroids"""

    def multiple_choice_from_dict(self, item_prompt: str, choices: dict[str, Any]) -> list[Any]:
        """Use to select multiple items from dictionary

        :param item_prompt: str:
        :param choices: dict[str, Any]:

        """
        selected: list[Any] = []
        dropdown = {"[DONE]": "done"} | choices
        while True:
            key = self.choice(item_prompt, list(dropdown.keys()))
            if key == "[DONE]":
                break
            selected.append(choices[key])
            del dropdown[key]
            if len(selected) == len(choices):
                # we've selected everything
                break
        return selected

    def choice_from_dict(self, text: str, choices: dict[str, Any], *, sort: bool = True) -> Any:
        """Use to select a value from the dictionary by showing users sorted dictionary keys

        :param text: str:
        :param choices: dict[str,Any]:
        :param *:
        :param sort: bool:  (Default value = True)

        """
        key = self.choice(text, list(choices.keys()), sort=sort)
        return choices[key]

    def choice(self, text: str, choices: list[Any], *, max_attempts: int = 10, sort: bool = True) -> str:
        """Use to select a value from a list

        :param text: str:
        :param choices: list[Any]:
        :param *:
        :param max_attempts: int:  (Default value = 10)
        :param sort: bool:  (Default value = True)

        """
        if sort:
            choices = sorted(choices, key=str.casefold)
        numbered = "\n".join(f"\033[1m[{i}]\033[0m \033[36m{v}\033[0m" for i, v in enumerate(choices))
        prompt = f"\033[1m{text}\033[0m\n{numbered}\nEnter a number between 0 and {len(choices) - 1}"
        attempt = 0
        while attempt < max_attempts:
            attempt += 1
            res = int(self.question(prompt, valid_number=True))
            if res >= len(choices) or res < 0:
                print(f"\033[31m[ERROR] Out of range: {res}\033[0m\n")
                continue
            return choices[res]
        msg = f"cannot get answer within {max_attempts} attempt"
        raise ValueError(msg)

    def confirm(self, text: str, *, max_attempts: int = 10):
        """Use to guard any optional or destructive actions of your app

        :param text: str:
        :param *:
        :param max_attempts: int:  (Default value = 10)

        """
        answer = self.question(text, valid_regex=r"[Yy][Ee][Ss]|[Nn][Oo]", default="no", max_attempts=max_attempts)
        return answer.lower() == "yes"

    def question(
        self,
        text: str,
        *,
        default: str | None = None,
        max_attempts: int = 10,
        valid_number: bool = False,
        valid_regex: str | None = None,
        validate: Callable[[str], bool] | None = None,
    ) -> str:
        """Use as testable alternative to `input()` builtin

        :param text: str:
        :param *:
        :param default: str | None:  (Default value = None)
        :param max_attempts: int:  (Default value = 10)
        :param valid_number: bool:  (Default value = False)
        :param valid_regex: str | None:  (Default value = None)
        :param validate: Callable[[str], bool] | None:  (Default value = None)

        """
        default_help = "" if default is None else f"\033[36m (default: {default})\033[0m"
        prompt = f"\033[1m{text}{default_help}: \033[0m"
        match_regex = None
        if valid_number:
            valid_regex = r"^\d+$"
        if valid_regex:
            match_regex = re.compile(valid_regex)
        attempt = 0
        while attempt < max_attempts:
            attempt += 1
            res = input(prompt)  # pylint: disable=bad-builtin
            if res and validate:
                if not validate(res):
                    continue
            if res and match_regex:
                if not match_regex.match(res):
                    print(f"\033[31m[ERROR] Not a '{valid_regex}' match: {res}\033[0m\n")
                    continue
                return res
            if not res and default:
                return default
            if not res:
                continue
            return res
        raise ValueError(f"cannot get answer within {max_attempts} attempt")


class MockPrompts(Prompts):
    """Testing utility for prompts"""

    def __init__(self, patterns_to_answers: dict[str, str]):
        patterns = [(re.compile(k), v) for k, v in patterns_to_answers.items()]
        self._questions_to_answers = sorted(patterns, key=lambda _: len(_[0].pattern), reverse=True)

    def question(self, text: str, default: str | None = None, **_) -> str:
        logger.info(f"Asking prompt: {text}")
        for question, answer in self._questions_to_answers:
            if not question.search(text):
                continue
            if not answer and default:
                return default
            return answer
        raise ValueError(f"not mocked: {text}")

    def extend(self, patterns_to_answers: dict[str, str]) -> MockPrompts:
        """Extend the existing list of questions and answers"""
        new_patterns_to_answers = {
            **{pattern.pattern: answer for pattern, answer in self._questions_to_answers},
            **patterns_to_answers,
        }
        return MockPrompts(new_patterns_to_answers)



"""This module provides a RateLimiter class and a rate_limited decorator to limit the rate of requests"""

import logging
import threading
import time
from functools import wraps

logger = logging.getLogger(__name__)


class RateLimiter:
    """Used to limit the rate of requests to a certain maximum rate within a specified burst period,
    defaulting to one second. It is nice to control and manage the flow of requests to prevent
    overloading a service.

    It uses a variation of the Leaky Bucket algorithm (https://en.wikipedia.org/wiki/Leaky_bucket)
    for its simplicity."""

    def __init__(self, *, max_requests: int = 30, burst_period_seconds: int = 1):
        self._capacity = max_requests
        self._burst_period_seconds = burst_period_seconds
        self._lock = threading.RLock()
        self._last = time.time()
        self._bucket = 0

    def throttle(self):
        """The throttle method is used to check and control the rate of requests. It's intended to be called
        before making a new request and is thread-safe.

        It calculates the time elapsed since the last request and subtracts this from the _burst_period_seconds
        to determine how much time remains in the current burst period. If the calculated delay is less than zero,
        it means that the burst period has elapsed, so it resets the request count and updates the last request
        timestamp. If the request count exceeds the allowed maximum (_max_requests), it sleeps for the remaining
        time in the burst period (delay) to enforce the rate limit."""
        with self._lock:
            now = time.time()
            delay = self._burst_period_seconds - (now - self._last)
            if delay < 0:
                # If the bucket is empty, it stops leaking.
                self._bucket = 0
                self._last = now
            self._bucket += 1
            if self._bucket > self._capacity:
                # If the bucket is over capacity, start leaking
                # by sleeping on the current thread.
                logger.debug(f"Throttled for {delay}s")
                time.sleep(delay)


def rate_limited(*, max_requests: int = 30, burst_period_seconds: int = 1):
    """Limits execution frequency of decorated function

    :param *: this function accepts keyword-only arguments
    :param max_requests: int:  (Default value = 30)
    :param burst_period_seconds: int:  (Default value = 1)

    """

    def decorator(func):
        rate_limiter = RateLimiter(max_requests=max_requests, burst_period_seconds=burst_period_seconds)

        @wraps(func)
        def wrapper(*args, **kwargs):
            rate_limiter.throttle()
            return func(*args, **kwargs)

        return wrapper

    return decorator



from __future__ import annotations

import abc
import builtins
import codecs
import fnmatch
import io
import locale
import logging
import os
import posixpath
import re
import shutil
import stat
from abc import abstractmethod
from collections.abc import Generator, Iterable, Sequence
from io import BytesIO, StringIO
from pathlib import Path, PurePath
from typing import NoReturn, TypeVar
from urllib.parse import quote_from_bytes as urlquote_from_bytes

from databricks.sdk import WorkspaceClient
from databricks.sdk.errors import DatabricksError, NotFound, ResourceDoesNotExist
from databricks.sdk.service.files import FileInfo
from databricks.sdk.service.workspace import (
    ExportFormat,
    ImportFormat,
    Language,
    ObjectInfo,
    ObjectType,
)

logger = logging.getLogger(__name__)


def _na(fn: str):
    def _inner(*_, **__):
        __tracebackhide__ = True  # pylint: disable=unused-variable
        raise NotImplementedError(f"{fn}() is not available for Databricks Workspace")

    return _inner


class _UploadIO(abc.ABC):
    def __init__(self, ws: WorkspaceClient, path: str):
        self._ws = ws
        self._path = path

    def close(self):
        # pylint: disable-next=no-member
        io_stream = self.getvalue()  # noqa
        self._ws.workspace.upload(self._path, io_stream, format=ImportFormat.AUTO)

    def __repr__(self):
        return f"<{self.__class__.__name__} for {self._path} on {self._ws}>"


class _BinaryUploadIO(_UploadIO, BytesIO):  # type: ignore
    def __init__(self, ws: WorkspaceClient, path: str):
        _UploadIO.__init__(self, ws, path)
        BytesIO.__init__(self)


class _TextUploadIO(_UploadIO, StringIO):  # type: ignore
    def __init__(self, ws: WorkspaceClient, path: str):
        _UploadIO.__init__(self, ws, path)
        StringIO.__init__(self)


P = TypeVar("P", bound="_DatabricksPath")


class _DatabricksPath(Path, abc.ABC):  # pylint: disable=too-many-public-methods
    """Base-class for experimental pathlib.Path implementations covering Databricks Workspace paths and DBFS."""

    # Implementation notes:
    #  - The builtin Path classes are not designed for extension, which in turn makes everything a little cumbersome.
    #  - The internals of the builtin pathlib have changed dramatically across supported Python versions (3.10-3.12 at
    #    the time of writing) so relying on those details is brittle. (Python 3.13 also includes a significant
    #    refactoring.)
    #  - Until 3.11 the implementation was decomposed and delegated to two internal interfaces:
    #     1. Flavour (scope=class) which encapsulates the path style and manipulation.
    #     2. Accessor (scope=instance) to which I/O-related calls are delegated.
    #    These interfaces are internal/protected.
    #  - Since 3.12 the implementation of these interfaces have been removed:
    #     1. Flavour has been replaced with posixpath and ntpath (normally imported as os.path). Still class-scoped.
    #     2. Accessor has been replaced with inline implementations based directly on the 'os' module.
    #
    # This implementation for Databricks-style paths does the following:
    #     1. Flavour is basically posix-style, with the caveat that we don't bother with the special //-prefix handling.
    #     2. The Accessor is replaced by delegation to existing routines available via the workspace client.
    #     3. Python 3.12 introduces some new API elements. Because these are source-compatible with earlier versions
    #        these are forward-ported and implemented.
    #
    # The current class hierarchy implements behaviour (and differentiation) by inheritance rather than composition.
    #
    __slots__ = (  # pylint: disable=redefined-slots-in-subclass
        # For us this is always the empty string. Consistent with the superclass attribute for Python 3.10-3.13b.
        "_drv",
        # The (normalized) root property for the path. Consistent with the superclass attribute for Python 3.10-3.13b.
        "_root",
        # The (normalized) path components (relative to the root) for the path.
        #  - For python <=3.11 this supersedes _parts
        #  - For python 3.12+ this supersedes _raw_paths
        "_path_parts",
        # The cached str() value of the instance. Consistent with the superclass attribute for Python 3.10-3.13b.
        "_str",
        # The cached hash() value for the instance. Consistent with the superclass attribute for Python 3.10-3.13b.
        "_hash",
        # The workspace client that we use to perform I/O operations on the path.
        "_ws",
    )
    _str: str
    _hash: int

    # Path semantics are posix-like.
    parser = posixpath

    # Compatibility attribute, for when superclass implementations get invoked on python <= 3.11.
    _flavour = object()

    # Public APIs that we don't support.
    as_uri = _na("as_uri")
    cwd = _na("cwd")
    chmod = _na("chmod")
    lchmod = _na("lchmod")
    lstat = _na("lstat")
    owner = _na("owner")
    group = _na("group")
    readlink = _na("readlink")
    symlink_to = _na("symlink_to")
    hardlink_to = _na("hardlink_to")
    touch = _na("touch")
    link_to = _na("link_to")
    samefile = _na("samefile")

    def __new__(cls, *args, **kwargs):
        # Force all initialisation to go via __init__() irrespective of the (Python-specific) base version.
        return object.__new__(cls)

    # pylint: disable=super-init-not-called
    def __init__(self, ws: WorkspaceClient, *args: str | bytes | os.PathLike) -> None:
        # We deliberately do _not_ call the super initializer because we're taking over complete responsibility for the
        # implementation of the public API.

        # Convert the arguments into string-based path segments, irrespective of their type.
        raw_paths = self._to_raw_paths(*args)

        # Normalise the paths that we have.
        root, path_parts = self._parse_and_normalize(raw_paths)
        self._drv = ""
        self._root = root
        self._path_parts = path_parts
        self._ws = ws

    @staticmethod
    def _to_raw_paths(*args: str | bytes | os.PathLike) -> list[str]:
        raw_paths: list[str] = []
        for arg in args:
            if isinstance(arg, PurePath):
                raw_paths.extend(arg.parts)
            else:
                try:
                    path = os.fspath(arg)
                except TypeError:
                    path = arg
                if not isinstance(path, str):
                    msg = (
                        f"argument should be a str or an os.PathLib object where __fspath__ returns a str, "
                        f"not {type(path).__name__!r}"
                    )
                    raise TypeError(msg)
                raw_paths.append(path)
        return raw_paths

    @classmethod
    def _parse_and_normalize(cls, parts: list[str]) -> tuple[str, tuple[str, ...]]:
        """Parse and normalize a list of path components.

        Args:
            parts: a list of path components to parse and normalize.
        Returns:
            A tuple containing:
              - The normalized drive (always '')
              - The normalized root for this path, or '' if there isn't any.
              - The normalized path components, if any, (relative) to the root.
        """
        match parts:
            case []:
                path = ""
            case [part]:
                path = part
            case [*parts]:
                path = cls.parser.join(*parts)
        if not path:
            return "", ()
        root, rel = cls._splitroot(path, sep=cls.parser.sep)
        # No need to split drv because we don't support it.
        parsed = tuple(str(x) for x in rel.split(cls.parser.sep) if x and x != ".")
        return root, parsed

    @classmethod
    def _splitroot(cls, part: str, sep: str) -> tuple[str, str]:
        # Based on the upstream implementation, with the '//'-specific bit elided because we don't need to
        # bother with Posix semantics.
        if part and part[0] == sep:
            return sep, part.lstrip(sep)
        return "", part

    @abstractmethod
    def as_fuse(self) -> Path: ...

    @abstractmethod
    def exists(self, *, follow_symlinks: bool = True) -> bool: ...

    @abstractmethod
    def _mkdir(self) -> None: ...

    @abstractmethod
    def rmdir(self, recursive: bool = False) -> None: ...

    @abstractmethod
    def unlink(self, missing_ok: bool = False) -> None: ...

    @abstractmethod
    def open(
        self,
        mode: str = "r",
        buffering: int = -1,
        encoding: str | None = None,
        errors: str | None = None,
        newline: str | None = None,
    ): ...

    @abstractmethod
    def is_dir(self) -> bool: ...

    @abstractmethod
    def is_file(self) -> bool: ...

    @abstractmethod
    def rename(self: P, target: str | bytes | os.PathLike) -> P: ...

    @abstractmethod
    def replace(self: P, target: str | bytes | os.PathLike) -> P: ...

    @abstractmethod
    def iterdir(self: P) -> Generator[P, None, None]: ...

    def __reduce__(self) -> NoReturn:
        # Cannot support pickling because we can't pickle the workspace client.
        msg = f"Pickling {self.__class__.__qualname__} paths is not supported."
        raise NotImplementedError(msg)

    def __fspath__(self):
        # Cannot support this: Workspace objects aren't accessible via the filesystem.
        #
        # This method is part of the os.PathLike protocol. Functions which accept a PathLike argument use os.fsname()
        # to convert (via this method) the object into a file system path that can be used with the low-level os.*
        # methods.
        #
        # Relevant online documentation:
        #  - PEP 519 (https://peps.python.org/pep-0519/)
        #  - os.fspath (https://docs.python.org/3/library/os.html#os.fspath)
        # TODO: Allow this to work when within an appropriate Databricks Runtime that mounts Workspace paths via FUSE.
        msg = f"{self.__class__.__qualname__} paths are not path-like: {self}"
        raise NotImplementedError(msg)

    def as_posix(self) -> str:
        return str(self)

    def __str__(self) -> str:
        try:
            return self._str
        except AttributeError:
            self._str = (self._root + self.parser.sep.join(self._path_parts)) or "."
            return self._str

    def __bytes__(self) -> bytes:
        return str(self).encode("utf-8")

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}({str(self)!r})"

    def __eq__(self, other) -> bool:
        if not isinstance(other, type(self)):
            return NotImplemented
        return str(self) == str(other)

    def __hash__(self) -> int:
        try:
            return self._hash
        except AttributeError:
            self._hash = hash(str(self))
            return self._hash

    @property
    def _parts(self) -> tuple[str, ...]:
        """Return a tuple that has the same natural ordering as paths of this type."""
        # Compatibility property (python <= 3.11), accessed via reverse equality comparison. This can't be avoided.
        return self._root, *self._path_parts

    _cparts = _parts

    @property
    def _str_normcase(self):
        # Compatibility property (python 3.12+), accessed via equality comparison. This can't be avoided.
        return str(self)

    @classmethod
    def _from_parts(cls, *args) -> NoReturn:
        # Compatibility method (python <= 3.11), accessed via reverse /-style building. This can't be avoided.
        # See __rtruediv__ for more information.
        raise TypeError("trigger NotImplemented")

    @property
    def _raw_paths(self) -> NoReturn:
        # Compatibility method (python 3.12+), accessed via reverse /-style building. This can't be avoided.
        # See __rtruediv__ for more information.
        raise TypeError("trigger NotImplemented")

    def __lt__(self, other) -> bool:
        if not isinstance(other, type(self)):
            return NotImplemented
        return self._path_parts < other._path_parts

    def __le__(self, other) -> bool:
        if not isinstance(other, type(self)):
            return NotImplemented
        return self._path_parts <= other._path_parts

    def __gt__(self, other) -> bool:
        if not isinstance(other, type(self)):
            return NotImplemented
        return self._path_parts > other._path_parts

    def __ge__(self, other) -> bool:
        if not isinstance(other, type(self)):
            return NotImplemented
        return self._path_parts >= other._path_parts

    def with_segments(self: P, *path_segments: bytes | str | os.PathLike) -> P:
        return type(self)(self._ws, *path_segments)

    @property
    def drive(self) -> str:
        return self._drv

    @property
    def root(self) -> str:
        return self._root

    @property
    def anchor(self) -> str:
        return self.drive + self.root

    @property
    def name(self) -> str:
        path_parts = self._path_parts
        return path_parts[-1] if path_parts else ""

    @property
    def parts(self) -> tuple[str, ...]:
        if self.drive or self.root:
            return self.drive + self.root, *self._path_parts
        return self._path_parts

    def with_name(self: P, name: str) -> P:
        parser = self.parser
        if not name or parser.sep in name or name == ".":
            msg = f"Invalid name: {name!r}"
            raise ValueError(msg)
        path_parts = list(self._path_parts)
        if not path_parts:
            raise ValueError(f"{self!r} has an empty name")
        path_parts[-1] = name
        return type(self)(self._ws, self.anchor, *path_parts)

    def with_suffix(self: P, suffix: str) -> P:
        stem = self.stem
        if not stem:
            msg = f"{self!r} has an empty name"
            raise ValueError(msg)
        if suffix and not suffix.startswith("."):
            msg = f"{self!r} invalid suffix: {suffix}"
            raise ValueError(msg)
        return self.with_name(stem + suffix)

    # pylint: disable=arguments-differ
    def relative_to(self: P, *other: str | bytes | os.PathLike, walk_up: bool = False) -> P:
        normalized = self.with_segments(*other)
        if self.anchor != normalized.anchor:
            msg = f"{str(self)!r} and {str(normalized)!r} have different anchors"
            raise ValueError(msg)
        path_parts0 = self._path_parts
        path_parts1 = normalized._path_parts  # pylint: disable=protected-access
        # Find the length of the common prefix.
        i = 0
        while i < len(path_parts0) and i < len(path_parts1) and path_parts0[i] == path_parts1[i]:
            i += 1
        relative_parts = path_parts0[i:]
        # Handle walking up.
        if i < len(path_parts1):
            if not walk_up:
                msg = f"{str(self)!r} is not in the subpath of {str(normalized)!r}"
                raise ValueError(msg)
            if ".." in path_parts1[i:]:
                raise ValueError(f"'..' segment in {str(normalized)!r} cannot be walked")
            walkup_parts = [".."] * (len(path_parts1) - i)
            relative_parts = (*walkup_parts, *relative_parts)
        return self.with_segments("", *relative_parts)

    def is_relative_to(self, *other: str | bytes | os.PathLike) -> bool:
        normalized = self.with_segments(*other)
        if self.anchor != normalized.anchor:
            return False
        path_parts0 = self._path_parts
        path_parts1 = normalized._path_parts  # pylint: disable=protected-access
        return path_parts0[: len(path_parts1)] == path_parts1

    @property
    def parent(self: P) -> P:
        rel_path = self._path_parts
        return self.with_segments(self.anchor, *rel_path[:-1]) if rel_path else self

    @property
    def parents(self: P) -> tuple[P, ...]:
        parents = []
        path = self
        parent = path.parent
        while path != parent:
            parents.append(parent)
            path = parent
            parent = path.parent
        return tuple(parents)

    def is_absolute(self) -> bool:
        return bool(self.anchor)

    def is_reserved(self) -> bool:
        return False

    def joinpath(self: P, *path_segments) -> P:
        return self.with_segments(self, *path_segments)

    def __truediv__(self: P, other: str | bytes | os.PathLike) -> P:
        try:
            return self.with_segments(*self._parts, other)
        except TypeError:
            return NotImplemented

    def __rtruediv__(self, other):
        # Note: this is only invoked if __truediv__ has already returned NotImplemented.
        # For the case of Path / _DatabricksPath this means the underlying __truediv__ is invoked.
        # The base-class implementations all access internals but yield NotImplemented if TypeError is raised. As
        # such we stub those internals (_from_parts and _raw_path) to trigger the NotImplemented path and ensure that
        # control ends up here.
        try:
            if isinstance(other, PurePath):
                return type(other)(other, *self._parts)
            return self.with_segments(other, *self._parts)
        except TypeError:
            return NotImplemented

    def match(self, path_pattern: str | bytes | os.PathLike, *, case_sensitive: bool | None = None) -> bool:
        # Convert the pattern to a fake path (with globs) to help with matching parts.
        if not isinstance(path_pattern, PurePath):
            path_pattern = self.with_segments(path_pattern)
        # Default to false if not specified.
        if case_sensitive is None:
            case_sensitive = True

        pattern_parts = path_pattern.parts
        if not pattern_parts:
            raise ValueError("empty pattern")
        # Short-circuit on situations where a match is logically impossible.
        path_parts = self.parts
        if len(path_parts) < len(pattern_parts) or len(path_parts) > len(pattern_parts) and path_pattern.anchor:
            return False
        # Check each part, starting from the end.
        for path_part, pattern_part in zip(reversed(path_parts), reversed(pattern_parts)):
            pattern = _PatternSelector.compile_pattern(pattern_part, case_sensitive=case_sensitive)
            if not pattern.match(path_part):
                return False
        return True

    def home(self):  # pylint: disable=arguments-differ
        """Return the user's home directory. Adapted from pathlib.Path"""
        return type(self)(self._ws, "~").expanduser()

    def expanduser(self: P) -> P:
        # Expand ~ (but NOT ~user) constructs.
        if not (self._drv or self._root) and self._path_parts and self._path_parts[0][:1] == "~":
            if self._path_parts[0] == "~":
                user_name = self._ws.current_user.me().user_name
            else:
                other_user = self._path_parts[0][1:]
                msg = f"Cannot determine home directory for: {other_user}"
                raise RuntimeError(msg)
            if user_name is None:
                raise RuntimeError("Could not determine home directory.")
            homedir = f"/Users/{user_name}"
            return self.with_segments(homedir, *self._path_parts[1:])
        return self

    def _return_false(self) -> bool:
        return False

    is_symlink = _return_false
    is_block_device = _return_false
    is_char_device = _return_false
    is_fifo = _return_false
    is_socket = _return_false
    is_mount = _return_false
    is_junction = _return_false

    def resolve(self: P, strict: bool = False) -> P:
        """Return the absolute path of the file or directory in Databricks Workspace."""
        absolute = self.absolute()
        if strict and not absolute.exists():
            msg = f"Path does not exist: {self}"
            raise FileNotFoundError(msg)
        return absolute

    def absolute(self: P) -> P:
        if self.is_absolute():
            return self
        return self.with_segments(self.cwd(), self)

    def mkdir(self, mode: int = 0o600, parents: bool = True, exist_ok: bool = True) -> None:
        """Create a directory;Only mode 0o600 is supported."""
        if not exist_ok:
            raise ValueError("exist_ok must be True for Databricks Workspace")
        if not parents:
            raise ValueError("parents must be True for Databricks Workspace")
        if mode != 0o600:
            raise ValueError("other modes than 0o600 are not yet supported")
        self._mkdir()

    def _prepare_pattern(self, pattern: str | bytes | os.PathLike) -> Sequence[str]:
        if not pattern:
            raise ValueError("Glob pattern must not be empty.")
        parsed_pattern = self.with_segments(pattern)
        if parsed_pattern.anchor:
            msg = f"Non-relative patterns are unsupported: {pattern!s}"
            raise NotImplementedError(msg)
        pattern_parts = parsed_pattern._path_parts  # pylint: disable=protected-access
        if ".." in pattern_parts:
            msg = f"Parent traversal is not supported: {pattern!s}"
            raise ValueError(msg)
        if os.fspath(pattern)[-1] == self.parser.sep:
            pattern_parts = (*pattern_parts, "")
        return pattern_parts

    def glob(
        self: P,
        pattern: str | bytes | os.PathLike,
        *,
        case_sensitive: bool | None = None,
    ) -> Generator[P, None, None]:
        pattern_parts = self._prepare_pattern(pattern)
        if case_sensitive is None:
            case_sensitive = True
        selector = _Selector.parse(pattern_parts, case_sensitive=case_sensitive)
        yield from selector(self)

    def rglob(
        self: P,
        pattern: str | bytes | os.PathLike,
        *,
        case_sensitive: bool | None = None,
    ) -> Generator[P, None, None]:
        pattern_parts = ("**", *self._prepare_pattern(pattern))
        if case_sensitive is None:
            case_sensitive = True
        selector = _Selector.parse(pattern_parts, case_sensitive=case_sensitive)
        yield from selector(self)


class DBFSPath(_DatabricksPath):
    """Experimental implementation of pathlib.Path for DBFS paths."""

    __slots__ = (
        # The cached _file_info value for the instance.
        "_cached_file_info",
    )
    _cached_file_info: FileInfo

    @classmethod
    def _from_file_info(cls, ws: WorkspaceClient, file_info: FileInfo) -> DBFSPath:
        """Special (internal-only) constructor that creates an instance based on FileInfo."""
        if not file_info.path:
            msg = f"Cannot initialise without file path: {file_info}"
            raise ValueError(msg)
        path = cls(ws, file_info.path)
        path._cached_file_info = file_info
        return path

    def as_fuse(self) -> Path:
        """Return FUSE-mounted path in Databricks Runtime."""
        if "DATABRICKS_RUNTIME_VERSION" not in os.environ:
            logger.warning("This method is only available in Databricks Runtime")
        return Path("/dbfs", self.as_posix().lstrip("/"))

    def exists(self, *, follow_symlinks: bool = True) -> bool:
        """Return True if the path points to an existing file, directory, or notebook"""
        if not follow_symlinks:
            raise NotImplementedError("follow_symlinks=False is not supported for DBFS")
        try:
            self._cached_file_info = self._ws.dbfs.get_status(self.as_posix())
            return True
        except NotFound:
            return False

    def _mkdir(self) -> None:
        self._ws.dbfs.mkdirs(self.as_posix())

    def rmdir(self, recursive: bool = False) -> None:
        """Remove a DBFS directory"""
        self._ws.dbfs.delete(self.as_posix(), recursive=recursive)

    def rename(self: P, target: str | bytes | os.PathLike) -> P:
        """Rename this path as the target, unless the target already exists."""
        dst = self.with_segments(target)
        self._ws.dbfs.move(self.as_posix(), dst.as_posix())
        return dst

    def replace(self: P, target: str | bytes | os.PathLike) -> P:
        """Rename this path, overwriting the target if it exists and can be overwritten."""
        dst = self.with_segments(target)
        if self.is_dir():
            msg = f"DBFS directories cannot currently be replaced: {self} -> {dst}"
            raise ValueError(msg)
        # Can't use self._ws.dbfs.move_(): it doesn't honour the overwrite flag properly.
        with dst.open(mode="wb") as writer, self.open(mode="rb") as reader:
            shutil.copyfileobj(reader, writer, length=1024 * 1024)
        self.unlink()
        return dst

    def unlink(self, missing_ok: bool = False) -> None:
        """Remove a file in Databricks Workspace."""
        # Although this introduces a race-condition, we have to handle missing_ok in advance because the DBFS client
        # doesn't report any error if deleting a target that doesn't exist.
        if not missing_ok and not self.exists():
            raise FileNotFoundError(f"{self.as_posix()} does not exist")
        self._ws.dbfs.delete(self.as_posix())

    def open(
        self,
        mode: str = "r",
        buffering: int = -1,
        encoding: str | None = None,
        errors: str | None = None,
        newline: str | None = None,
    ):
        """Open a DBFS file.

        Only text and binary I/O are supported in basic read or write mode, along with 'x' to avoid overwriting."""
        is_write = "w" in mode
        is_read = "r" in mode or not is_write
        if is_read and is_write:
            msg = f"Unsupported mode: {mode} (simultaneous read and write)"
            raise ValueError(msg)
        is_binary = "b" in mode
        is_text = "t" in mode or not is_binary
        if is_binary and is_text:
            msg = f"Unsupported mode: {mode} (binary and text I/O)"
            raise ValueError(msg)
        is_overwrite = is_write and "x" not in mode
        binary_io = self._ws.dbfs.open(self.as_posix(), read=is_read, write=is_write, overwrite=is_overwrite)
        if is_text:
            return io.TextIOWrapper(binary_io, encoding=encoding, errors=errors, newline=newline)
        return binary_io

    def write_bytes(self, data):
        """Write the (binary) data to this path."""
        # The DBFS BinaryIO implementation only accepts bytes and rejects the (other) byte-like builtins.
        match data:
            case builtins.bytes | builtins.bytearray:
                binary_data = bytes(data)
            case _:
                binary_data = bytes(memoryview(data))
        with self.open("wb") as f:
            return f.write(binary_data)

    @property
    def _file_info(self) -> FileInfo:
        # this method is cached because it is used in multiple is_* methods.
        # DO NOT use this method in methods, where fresh result is required.
        try:
            return self._cached_file_info
        except AttributeError:
            self._cached_file_info = self._ws.dbfs.get_status(self.as_posix())
            return self._cached_file_info

    def stat(self, *, follow_symlinks=True) -> os.stat_result:
        seq: list[float] = [-1.0] * 10
        seq[stat.ST_SIZE] = self._file_info.file_size or -1  # 6
        seq[stat.ST_MTIME] = (
            float(self._file_info.modification_time) / 1000.0 if self._file_info.modification_time else -1.0
        )  # 8
        return os.stat_result(seq)

    def is_dir(self) -> bool:
        """Return True if the path points to a DBFS directory."""
        try:
            return bool(self._file_info.is_dir)
        except DatabricksError:
            return False

    def is_file(self) -> bool:
        """Return True if the path points to a file in Databricks Workspace."""
        return not self.is_dir()

    def iterdir(self) -> Generator[DBFSPath, None, None]:
        for child in self._ws.dbfs.list(self.as_posix()):
            yield self._from_file_info(self._ws, child)


class WorkspacePath(_DatabricksPath):
    """Experimental implementation of pathlib.Path for Databricks Workspace."""

    __slots__ = (
        # The cached _object_info value for the instance.
        "_cached_object_info",
    )
    _cached_object_info: ObjectInfo

    _SUFFIXES = {".py": Language.PYTHON, ".sql": Language.SQL, ".scala": Language.SCALA, ".R": Language.R}

    @classmethod
    def _from_object_info(cls, ws: WorkspaceClient, object_info: ObjectInfo) -> WorkspacePath:
        """Special (internal-only) constructor that creates an instance based on ObjectInfo."""
        if not object_info.path:
            msg = f"Cannot initialise without object path: {object_info}"
            raise ValueError(msg)
        path = cls(ws, object_info.path)
        path._cached_object_info = object_info
        return path

    def as_uri(self) -> str:
        return f"{self._ws.config.host}#workspace{urlquote_from_bytes(bytes(self))}"

    def as_fuse(self) -> Path:
        """Return FUSE-mounted path in Databricks Runtime."""
        if "DATABRICKS_RUNTIME_VERSION" not in os.environ:
            logger.warning("This method is only available in Databricks Runtime")
        return Path("/Workspace", self.as_posix().lstrip("/"))

    def exists(self, *, follow_symlinks: bool = True) -> bool:
        """Return True if the path points to an existing file, directory, or notebook"""
        if not follow_symlinks:
            raise NotImplementedError("follow_symlinks=False is not supported for Databricks Workspace")
        try:
            self._cached_object_info = self._ws.workspace.get_status(self.as_posix())
            return True
        except NotFound:
            return False

    def _mkdir(self) -> None:
        self._ws.workspace.mkdirs(self.as_posix())

    def rmdir(self, recursive: bool = False) -> None:
        """Remove a directory in Databricks Workspace"""
        self._ws.workspace.delete(self.as_posix(), recursive=recursive)

    def _rename(self: P, target: str | bytes | os.PathLike, overwrite: bool) -> P:
        """Rename a file in Databricks Workspace"""
        dst = self.with_segments(target)
        if self.is_dir():
            msg = f"Workspace directories cannot currently be renamed: {self} -> {dst}"
            raise ValueError(msg)
        with self._ws.workspace.download(self.as_posix(), format=ExportFormat.AUTO) as f:
            self._ws.workspace.upload(dst.as_posix(), f.read(), format=ImportFormat.AUTO, overwrite=overwrite)
        self.unlink()
        return dst

    def rename(self, target: str | bytes | os.PathLike):
        """Rename this path as the target, unless the target already exists."""
        return self._rename(target, overwrite=False)

    def replace(self, target: str | bytes | os.PathLike):
        """Rename this path, overwriting the target if it exists and can be overwritten."""
        return self._rename(target, overwrite=True)

    def unlink(self, missing_ok: bool = False) -> None:
        """Remove a file in Databricks Workspace."""
        try:
            self._ws.workspace.delete(self.as_posix())
        except ResourceDoesNotExist as e:
            if not missing_ok:
                raise FileNotFoundError(f"{self.as_posix()} does not exist") from e

    def open(
        self,
        mode: str = "r",
        buffering: int = -1,
        encoding: str | None = None,
        errors: str | None = None,
        newline: str | None = None,
    ):
        """Open a file in Databricks Workspace. Only text and binary modes are supported."""
        if "b" in mode and "r" in mode:
            return self._ws.workspace.download(self.as_posix(), format=ExportFormat.AUTO)
        if "b" in mode and "w" in mode:
            return _BinaryUploadIO(self._ws, self.as_posix())
        if "r" in mode:
            with self._ws.workspace.download(self.as_posix(), format=ExportFormat.AUTO) as f:
                data = f.read()
                if encoding is None:
                    if data.startswith(codecs.BOM_UTF32_LE) or data.startswith(codecs.BOM_UTF32_BE):
                        encoding = "utf-32"
                    elif data.startswith(codecs.BOM_UTF16_LE) or data.startswith(codecs.BOM_UTF16_BE):
                        encoding = "utf-16"
                    elif data.startswith(codecs.BOM_UTF8):
                        encoding = "utf-8-sig"
                if encoding is None or encoding == "locale":
                    encoding = locale.getpreferredencoding(False)
                return StringIO(data.decode(encoding))
        if "w" in mode:
            return _TextUploadIO(self._ws, self.as_posix())
        raise ValueError(f"invalid mode: {mode}")

    def read_text(self, encoding=None, errors=None):
        with self.open(mode="r", encoding=encoding, errors=errors) as f:
            return f.read()

    @property
    def suffix(self) -> str:
        """Return the file extension. If the file is a notebook, return the suffix based on the language."""
        suffix = super().suffix
        if suffix:
            return suffix
        if not self.is_notebook():
            return ""
        for sfx, lang in self._SUFFIXES.items():
            try:
                if self._object_info.language == lang:
                    return sfx
            except DatabricksError:
                return ""
        return ""

    @property
    def _object_info(self) -> ObjectInfo:
        # this method is cached because it is used in multiple is_* methods.
        # DO NOT use this method in methods, where fresh result is required.
        try:
            return self._cached_object_info
        except AttributeError:
            self._cached_object_info = self._ws.workspace.get_status(self.as_posix())
            return self._object_info

    def stat(self, *, follow_symlinks=True) -> os.stat_result:
        seq: list[float] = [-1.0] * 10
        seq[stat.ST_SIZE] = self._object_info.size or -1  # 6
        seq[stat.ST_MTIME] = (
            float(self._object_info.modified_at) / 1000.0 if self._object_info.modified_at else -1.0
        )  # 8
        seq[stat.ST_CTIME] = float(self._object_info.created_at) / 1000.0 if self._object_info.created_at else -1.0  # 9
        return os.stat_result(seq)

    def is_dir(self) -> bool:
        """Return True if the path points to a directory in Databricks Workspace."""
        try:
            return self._object_info.object_type == ObjectType.DIRECTORY
        except DatabricksError:
            return False

    def is_file(self) -> bool:
        """Return True if the path points to a file in Databricks Workspace."""
        try:
            return self._object_info.object_type == ObjectType.FILE
        except DatabricksError:
            return False

    def is_notebook(self) -> bool:
        """Return True if the path points to a notebook in Databricks Workspace."""
        try:
            return self._object_info.object_type == ObjectType.NOTEBOOK
        except DatabricksError:
            return False

    def iterdir(self) -> Generator[WorkspacePath, None, None]:
        for child in self._ws.workspace.list(self.as_posix()):
            yield self._from_object_info(self._ws, child)


T = TypeVar("T", bound="Path")


class _Selector(abc.ABC):
    @classmethod
    def parse(cls, pattern_parts: Sequence[str], *, case_sensitive: bool) -> _Selector:
        # The pattern language is:
        #  - '**' matches any number (including zero) of file or directory segments. Must be the entire segment.
        #  - '*' match any number of characters within a single segment.
        #  - '?' match a single character within a segment.
        #  - '[seq]' match a single character against the class (within a segment).
        #  - '[!seq]' negative match for a single character against the class (within a segment).
        #  - A trailing '/' (which presents here as a trailing empty segment) matches only directories.
        # There is no explicit escaping mechanism; literal matches against special characters above are possible as
        # character classes, for example: [*]
        #
        # Some sharp edges:
        #  - Multiple '**' segments are allowed.
        #  - Normally the '..' segment is allowed. (This can be used to match against siblings, for
        #    example: /home/bob/../jane/) However WorspacePath (and DBFS) do not support '..' traversal in paths.
        #  - Normally '.' is allowed, but eliminated before we reach this method.
        match pattern_parts:
            case ["**", *tail]:
                return _RecursivePatternSelector(tail, case_sensitive=case_sensitive)
            case [head, *tail] if case_sensitive and not _PatternSelector.needs_pattern(head):
                return _LiteralSelector(head, tail, case_sensitive=case_sensitive)
            case [head, *tail]:
                if "**" in head:
                    raise ValueError("Invalid pattern: '**' can only be a complete path component")
                return _PatternSelector(head, tail, case_sensitive=case_sensitive)
            case []:
                return _TerminalSelector()
        raise ValueError(f"Glob pattern unsupported: {pattern_parts}")

    @abstractmethod
    def __call__(self, path: T) -> Iterable[T]: ...


class _TerminalSelector(_Selector):
    def __call__(self, path: T) -> Iterable[T]:
        yield path


class _NonTerminalSelector(_Selector):
    __slots__ = (
        "_dir_only",
        "_child_selector",
    )

    def __init__(self, child_pattern_parts: Sequence[str], *, case_sensitive: bool) -> None:
        super().__init__()
        if child_pattern_parts:
            self._child_selector = self.parse(child_pattern_parts, case_sensitive=case_sensitive)
            self._dir_only = True
        else:
            self._child_selector = _TerminalSelector()
            self._dir_only = False

    def __call__(self, path: T) -> Iterable[T]:
        if path.is_dir():
            yield from self._select_children(path)

    @abstractmethod
    def _select_children(self, path: T) -> Iterable[T]: ...


class _LiteralSelector(_NonTerminalSelector):
    __slots__ = ("_literal_path",)

    def __init__(self, path: str, child_pattern_parts: Sequence[str], case_sensitive: bool) -> None:
        super().__init__(child_pattern_parts, case_sensitive=case_sensitive)
        self._literal_path = path

    def _select_children(self, path: T) -> Iterable[T]:
        candidate = path / self._literal_path
        if self._dir_only and candidate.is_dir() or candidate.exists():
            yield from self._child_selector(candidate)


class _PatternSelector(_NonTerminalSelector):
    __slots__ = ("_pattern",)

    # The special set of characters that indicate a glob pattern isn't a trivial literal.
    # Ref: https://docs.python.org/3/library/fnmatch.html#module-fnmatch
    _glob_specials = re.compile("[*?\\[\\]]")

    @classmethod
    def needs_pattern(cls, pattern: str) -> bool:
        return cls._glob_specials.search(pattern) is not None

    @classmethod
    def compile_pattern(cls, pattern: str, case_sensitive: bool) -> re.Pattern:
        flags = 0 if case_sensitive else re.IGNORECASE
        regex = fnmatch.translate(pattern)
        return re.compile(regex, flags=flags)

    def __init__(self, pattern: str, child_pattern_parts: Sequence[str], case_sensitive: bool) -> None:
        super().__init__(child_pattern_parts, case_sensitive=case_sensitive)
        self._pattern = self.compile_pattern(pattern, case_sensitive=case_sensitive)

    def _select_children(self, path: T) -> Iterable[T]:
        candidates = list(path.iterdir())
        for candidate in candidates:
            if self._dir_only and not candidate.is_dir():
                continue
            if self._pattern.match(candidate.name):
                yield from self._child_selector(candidate)


class _RecursivePatternSelector(_NonTerminalSelector):
    def __init__(self, child_pattern_parts: Sequence[str], case_sensitive: bool) -> None:
        super().__init__(child_pattern_parts, case_sensitive=case_sensitive)

    def _all_directories(self, path: T) -> Iterable[T]:
        # Depth-first traversal of directory tree, visiting this node first.
        yield path
        children = [child for child in path.iterdir() if child.is_dir()]
        for child in children:
            yield from self._all_directories(child)

    def _select_children(self, path: T) -> Iterable[T]:
        yielded = set()
        for starting_point in self._all_directories(path):
            for candidate in self._child_selector(starting_point):
                if candidate not in yielded:
                    yielded.add(candidate)
                    yield candidate



import sys
from pathlib import Path

from databricks.labs.blueprint.cli import App
from databricks.labs.blueprint.entrypoint import (
    find_project_root,
    get_logger,
    relative_paths,
)
from databricks.labs.blueprint.tui import Prompts

blueprint = App(__file__)
logger = get_logger(__file__)

MAIN_PY_FILE = '''from databricks.sdk import AccountClient, WorkspaceClient
from databricks.labs.blueprint.entrypoint import get_logger
from databricks.labs.blueprint.cli import App

__app__ = App(__file__)
logger = get_logger(__file__)


@__app__.command
def me(w: WorkspaceClient, greeting: str):
    """Shows current username"""
    logger.info(f"{greeting}, {w.current_user.me().user_name}!")


@__app__.command(is_account=True)
def workspaces(a: AccountClient):
    """Shows workspaces"""
    for ws in a.workspaces.list():
        logger.info(f"Workspace: {ws.workspace_name} ({ws.workspace_id})")


if "__main__" == __name__:
    __app__()

'''

LABS_YML_FILE = """---
name: __app__
description: Common libraries for Databricks Labs
install:
  script: src/databricks/labs/__app__/__init__.py
entrypoint: src/databricks/labs/__app__/__main__.py
min_python: 3.10
  - name: me
    description: shows current username
    flags:
     - name: greeting
       default: Hello
       description: Greeting prefix
  - name: workspaces
    is_account: true
    description: shows current workspaces
"""


@blueprint.command(is_unauthenticated=True)
def init_project(target):
    """Creates the required boilerplate structure"""
    prompts = Prompts()

    project_root = find_project_root(__file__)
    target_folder = Path(target)

    project_name = prompts.question("Name of the project", default=target_folder.name)
    src_dir, dst_dir = relative_paths(project_root, target_folder.absolute())

    ignore_names = {
        ".git",
        ".venv",
        ".databricks",
        ".mypy_cache",
        ".idea",
        ".coverage",
        "htmlcov",
        "__pycache__",
        "tests",
        ".databricks-login.json",
        "coverage.xml",
        "dist",
        "docs",
    }
    queue: list[Path] = [src_dir]  # type: ignore[annotation-unchecked]
    while queue:
        current = queue.pop(0)
        if current.name in ignore_names:
            continue
        if current.is_file():
            relative_file_name = current.as_posix().replace("blueprint", project_name)
            dst_file = dst_dir / relative_file_name
            dst_file.parent.mkdir(exist_ok=True, parents=True)
            with current.open("r", encoding=sys.getdefaultencoding()) as src, dst_file.open("w") as dst:
                content = src.read().replace("blueprint", project_name)
                content = content.replace("databricks-sdk", "databricks-labs-blueprint")
                dst.write(content)
            continue
        virtual_env_marker = current / "pyvenv.cfg"
        if virtual_env_marker.exists():
            continue
        for file in current.iterdir():
            if file.as_posix() == "src/databricks/labs/blueprint":
                continue
            queue.append(file)
    inner_package_dir = dst_dir / "src" / "databricks" / "labs" / project_name
    inner_package_dir.mkdir(parents=True, exist_ok=True)
    with (inner_package_dir / "__main__.py").open("w") as f:
        f.write(MAIN_PY_FILE.replace("__app__", project_name))
    with (inner_package_dir / "__init__.py").open("w") as f:
        f.write(f"from databricks.labs.{project_name}.__about__ import __version__")
    with (inner_package_dir / "__about__.py").open("w") as f:
        f.write('__version__ = "0.0.0"\n')
    with (dst_dir / "labs.yml").open("w") as f:
        f.write(LABS_YML_FILE.replace("__app__", project_name))
    with (dst_dir / "CODEOWNERS").open("w") as f:
        f.write(f"* @nfx\n/src @databrickslabs/{project_name}-write\n/tests @databrickslabs/{project_name}-write\n")
    with (dst_dir / "CHANGELOG.md").open("w") as f:
        f.write(f"# Version changelog\n\n## 0.0.0\n\nInitial {project_name} commit\n")


if __name__ == "__main__":
    blueprint()



"""A nice formatter for logging. It uses colors and bold text if the console supports it."""

import logging
import sys


class NiceFormatter(logging.Formatter):
    """A nice formatter for logging. It uses colors and bold text if the console supports it."""

    BOLD = "\033[1m"
    RESET = "\033[0m"
    GREEN = "\033[32m"
    BLACK = "\033[30m"
    CYAN = "\033[36m"
    YELLOW = "\033[33m"
    RED = "\033[31m"
    MAGENTA = "\033[35m"
    GRAY = "\033[90m"

    def __init__(self, *, probe_tty: bool = False) -> None:
        """Create a new instance of the formatter. If probe_tty is True, then the formatter will
        attempt to detect if the console supports colors. If probe_tty is False, colors will be
        enabled by default."""
        super().__init__(fmt="%(asctime)s %(levelname)s [%(name)s] %(message)s", datefmt="%H:%M")
        self._levels = {
            logging.NOTSET: self._bold("TRACE"),
            logging.DEBUG: self._bold(f"{self.CYAN}DEBUG"),
            logging.INFO: self._bold(f"{self.GREEN} INFO"),
            logging.WARNING: self._bold(f"{self.YELLOW} WARN"),
            logging.ERROR: self._bold(f"{self.RED}ERROR"),
            logging.CRITICAL: self._bold(f"{self.MAGENTA}FATAL"),
        }
        # show colors in runtime, github actions, and while debugging
        self.colors = sys.stdout.isatty() if probe_tty else True

    def _bold(self, text):
        """Return text in bold."""
        return f"{self.BOLD}{text}{self.RESET}"

    def format(self, record: logging.LogRecord):  # noqa: A003
        """Format the log record. If colors are enabled, use them."""
        if not self.colors:
            return super().format(record)
        timestamp = self.formatTime(record, datefmt="%H:%M:%S")
        level = self._levels[record.levelno]
        # databricks.labs.ucx.foo.bar -> d.l.u.foo.bar
        module_split = record.name.split(".")
        last_two_modules = len(module_split) - 2
        name = ".".join(part if i >= last_two_modules else part[0] for i, part in enumerate(module_split))
        msg = record.msg
        if record.exc_info and not record.exc_text:
            record.exc_text = self.formatException(record.exc_info)
        if record.exc_text:
            msg += ": " + record.exc_text
        if record.stack_info:
            if msg[-1:] != "\n":
                msg += "\n"
            msg += self.formatStack(record.stack_info)

        color_marker = self.GRAY
        if record.levelno in (logging.INFO, logging.WARNING):
            color_marker = self.BOLD
        elif record.levelno in (logging.ERROR, logging.FATAL):
            color_marker = self.RED + self.BOLD
        thread_name = ""
        if record.threadName != "MainThread":
            thread_name = f"[{record.threadName}]"
        return f"{self.GRAY}{timestamp}{self.RESET} {level} {color_marker}[{name}]{thread_name} {msg}{self.RESET}"


def install_logger(level="DEBUG"):
    """Install a console logger with a nice formatter."""
    for handler in logging.root.handlers:
        logging.root.removeHandler(handler)
    console_handler = logging.StreamHandler(sys.stderr)
    console_handler.setFormatter(NiceFormatter())
    console_handler.setLevel(level)
    logging.root.addHandler(console_handler)
    return console_handler



"""Run tasks in parallel and return results and errors"""

import concurrent
import datetime as dt
import functools
import logging
import os
import re
import threading
from collections.abc import Callable, Collection, Sequence
from concurrent.futures import ThreadPoolExecutor
from typing import Generic, TypeVar

MIN_THREADS = 8

Result = TypeVar("Result")
Task = Callable[[], Result | None] | functools.partial[Result | None]
logger = logging.getLogger(__name__)


class ManyError(RuntimeError):
    def __init__(self, errs: Sequence[BaseException]):
        strs = sorted({f"{_.__class__.__name__}: {_!s}" for _ in errs})
        msg = f'Detected {len(errs)} failures: {", ".join(strs)}'
        super().__init__(msg)
        self.errs = errs


class Threads(Generic[Result]):
    def __init__(self, name, tasks: Sequence[Task[Result]], num_threads: int):
        self._name = name
        self._tasks = list(tasks)
        self._task_fail_error_pct = 50
        self._num_threads = num_threads
        self._started = dt.datetime.now()
        self._lock = threading.Lock()
        self._completed_cnt = 0
        self._large_log_every = 3000
        self._default_log_every = 100

    @classmethod
    def gather(
        cls, name: str, tasks: Sequence[Task[Result]], num_threads: int | None = None
    ) -> tuple[Collection[Result], list[Exception]]:
        """Run tasks in parallel and return results and errors"""
        if num_threads is None:
            num_cpus = os.cpu_count()
            if num_cpus is None:
                num_cpus = 1
            num_threads = max(num_cpus * 2, MIN_THREADS)
        return cls(name, tasks, num_threads=num_threads)._run()

    @classmethod
    def strict(cls, name: str, tasks: Sequence[Task[Result]]) -> Collection[Result]:
        """Run tasks in parallel and raise ManyError if any task fails"""
        # this dunder variable is hiding this method from tracebacks, making it cleaner
        # for the user to see the actual error without too much noise.
        __tracebackhide__ = True  # pylint: disable=unused-variable
        collected, errs = cls.gather(name, tasks)
        if errs:
            if len(errs) == 1:
                raise errs[0]
            raise ManyError(errs)
        return collected

    def _run(self) -> tuple[Collection[Result], list[Exception]]:
        """Run tasks in parallel and return results and errors"""
        given_cnt = len(self._tasks)
        if given_cnt == 0:
            return [], []
        logger.debug(f"Starting {given_cnt} tasks in {self._num_threads} threads")

        collected = []
        errors = []
        for future in self._execute():
            return_value = future.result()
            if return_value is None:
                continue
            result, err = return_value
            if err is not None:
                errors.append(err)
                continue
            if result is None:
                continue
            collected.append(result)
        self._on_finish(given_cnt, len(collected), len(errors))

        return collected, errors

    def _on_finish(self, given_cnt: int, collected_cnt: int, failed_cnt: int):
        """Log the results of the parallel execution"""
        since = dt.datetime.now() - self._started
        success_pct = collected_cnt / given_cnt * 100
        stats = f"{success_pct:.0f}% results available ({collected_cnt}/{given_cnt}). Took {since}"
        if failed_cnt == given_cnt:
            logger.critical(f"All '{self._name}' tasks failed!!!")
        elif failed_cnt > 0 and success_pct <= self._task_fail_error_pct:
            logger.error(f"More than half '{self._name}' tasks failed: {stats}")
        elif failed_cnt > 0:
            logger.warning(f"Some '{self._name}' tasks failed: {stats}")
        else:
            logger.info(f"Finished '{self._name}' tasks: {stats}")

    def _execute(self):
        """Run tasks in parallel and return futures"""
        thread_name_prefix = re.sub(r"\W+", "_", self._name)
        with ThreadPoolExecutor(self._num_threads, thread_name_prefix) as pool:
            futures = []
            for task in self._tasks:
                if task is None:
                    continue
                future = pool.submit(self._wrap_result(task, self._name))
                future.add_done_callback(self._progress_report)
                futures.append(future)
            return concurrent.futures.as_completed(futures)

    def _progress_report(self, _):
        """Log the progress of the parallel execution"""
        total_cnt = len(self._tasks)
        log_every = self._default_log_every
        if total_cnt > self._large_log_every:
            log_every = 500
        elif total_cnt <= self._default_log_every:
            log_every = 10
        with self._lock:
            self._completed_cnt += 1
            since = dt.datetime.now() - self._started
            rps = self._completed_cnt / since.total_seconds()
            if self._completed_cnt % log_every == 0 or self._completed_cnt == total_cnt:
                msg = f"{self._name} {self._completed_cnt}/{total_cnt}, rps: {rps:.3f}/sec"
                logger.info(msg)

    @staticmethod
    def _get_result_function_signature(func, name):
        if not isinstance(func, functools.partial):
            return name

        # try to build up signature, this should never fail
        try:
            args = []
            args.extend(repr(x) for x in func.args)
            args.extend(f"{k}={v!r}" for (k, v) in func.keywords.items())
            args_str = ", ".join(args)
            if args_str:
                return f"{name}({args_str})"
            return name
        # but if it would ever fail, better return generic serialized name, than messing up traceback even more...
        except Exception:  # pylint: disable=broad-exception-caught
            return str(func)

    @classmethod
    def _wrap_result(cls, func, name):
        """This method emulates GoLang's error return style"""

        @functools.wraps(func)
        def inner(*args, **kwargs):
            try:
                return func(*args, **kwargs), None
            except Exception as err:  # pylint: disable=broad-exception-caught
                signature = cls._get_result_function_signature(func, name)
                logger.error(f"{signature} task failed: {err!s}", exc_info=err)
                return None, err

        return inner
